#!/usr/bin/env python
# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.
#
# -*- coding: UTF-8 -*-
#
# ecore2smd - script to convert ECORE metamodels to SMD metamodels
#
# SYNOPSIS
#
# Explained in usage()
#
# DESCRIPTION
#
# ecore2smd is a script for converting ECORE metamodels (i.e. Eclipse)
# to SMD metamodels.
#
# BUGS
#
# Probably, the ecore metamodel seems to be changing
#
# AUTHORS
#
# Marcus Alanen <maalanen@abo.fi> 2005, 2006
#

coral = None
core = None
ecore_model = None
ecore_model_skip_these_datatypes = None
smd = None

def c(name):
    return core.elements.select(lambda x: x.name == name)[0]

def is_a_redefinition(e):
    # Skip if this is a redefinition?
    for annotation in e.eAnnotations:
        if annotation.source == "redefines":
            return True
    return False

def get_non_redefined_supersets(ss):
    todo = dict.fromkeys(ss)
    res = {}
    while todo:
        k = todo.keys()
        todo = {}
        for s in k:
            if res.has_key(s):
                continue
            if is_a_redefinition(s) or (s.eOpposite and is_a_redefinition(s.eOpposite)):
                for annotation in s.eAnnotations:
                    if annotation.source == "subsets":
                        for as in annotation.references:
                            todo[as] = None
            else:
                res[s] = None
    return res.keys()

def create(e):
    global ecore_model_skip_these_datatypes
    n = None
    s=None
    if e.oclIsKindOf(ecore.EPackage):
        s = smd.Language()
        s.URInamespace = e.nsURI
        s.version = 1
        s.revision = 0

    if e.oclIsKindOf(ecore.EEnum):
        s = smd.EnumerationDefinition()
        for i in e.eLiterals:
            lit = coral.String.newAtom()
            lit.__setString__(unicode(i.name))
            s.values.append(lit)
    elif e.oclIsKindOf(ecore.EDataType):
        return None
    
    if e.oclIsKindOf(ecore.EClass):
        s = smd.ElementDefinition()
        s.isAbstract = e.interface

    if e.oclIsKindOf(ecore.EReference):
        if is_a_redefinition(e):
            return None
        if e.eOpposite and is_a_redefinition(e.eOpposite):
            return None
        
    if e.oclIsKindOf(ecore.EStructuralFeature):
        if e.eType in ecore_model_skip_these_datatypes:
            return None

        if is_a_redefinition(e):
            return None
            
        s = smd.Property()
        s.isBag = e.unique
        s.multiplicity_lower = e.lowerBound
        s.multiplicity_upper = e.upperBound
        if e.transient:
            s.isUnserializable = "true"
        s.defaultValue = e.defaultValueLiteral

        s.isOrdered = e.ordered

        s.isDerived = e.derived
        for annotation in e.eAnnotations:
            if annotation.source == "union":
                s.isDerived = "true"
                break
            
        # ignore unsettable, we don't have it
        # ignore many, it's not useful
        # ignore required, it's not useful
        # ignore defaultValue, we don't have it

    if e.oclIsKindOf(ecore.EAttribute):
        s.type = "composition"
    if e.oclIsKindOf(ecore.EReference):
        if e.containment:
            s.type = "composition"
        # ignore container, it's useless
        # ignore resolveProxies, we don't have it

    if s and e.oclIsKindOf(ecore.ENamedElement):
        s.name = e.name
    return s

def connect(e2s):
    for e, s in e2s.iteritems():
        if e.oclIsKindOf(ecore.EPackage):
            for child in e.eClassifiers:
                # Skip the ignored stuff
                if e2s.has_key(child):
                    s.elements.add(e2s[child])
        if e.oclIsKindOf(ecore.EClass):
            for child in e.eStructuralFeatures:
                # Skip the ignored stuff
                if e2s.has_key(child):
                    s.properties.add(e2s[child])
            for annotation in e.eAnnotations:
                if annotation.source == "duplicates":
                    for child in annotation.contents:
                        # Skip the ignored stuff
                        if e2s.has_key(child):
                            s.properties.add(e2s[child])
            # Superclasses
            for superclass in e.eSuperTypes:
                # Skip those pointing to ecore_model
                if superclass.__getRoot__() == ecore_model:
                    continue
                s.superclasses.add(e2s[superclass])
            if not s.superclasses:
                s.superclasses.add(core.AnyElement)

        if e.oclIsKindOf(ecore.EReference):
            if e.eOpposite:
                s.opposite = e2s[e.eOpposite]
            for annotation in e.eAnnotations:
                if annotation.source == "subsets":
                    #
                    # We might subset a redefined property.
                    # We don't want redefinitions, so
                    # instead subset transitively the supersets
                    # of the redefined property.
                    #
                    lst = get_non_redefined_supersets(annotation.references)
                    for l in lst:
                        s.supersets.add(e2s[l])

        if e.oclIsKindOf(ecore.EDataType):
            continue

        if e.oclIsKindOf(ecore.EStructuralFeature):
            if e.eType in ecore_model_skip_these_datatypes:
                return
            # eAttributeType == eType
            h = {
                "Boolean": c("Boolean"),
                "EBigDecimal": c("Integer"),
                "EBigInteger": c("Integer"),
                "EBoolean": c("Boolean"),
                "EDouble": c("Double"),
                "EInt": c("Integer"),
                "EInteger": c("Integer"),
                "EJavaClass": c("coralPyObj"),
                "EJavaObject": c("coralPyObj"),
                "ELong": c("Integer"),
                "EShort": c("Integer"),
                "EString": c("String"),
                "Integer": c("Integer"),
                "String": c("String"),
                "UnlimitedNatural": c("Integer"),
                }
            try:
                s.element_kind = h[str(e.eType.name)]
            except KeyError:
                s.element_kind = e2s[e.eType]

def main(arguments):
    global ecore
    global ecore_model
    global smd
    global core
    global coral
    global ecore_model_skip_these_datatypes

    def usage():
        print "Usage: ecore2smd [ --help | --version | [ -o smdfile ] ecorefile"
        print "Options:"
        print "  --help         Display this information"
        print "  --version      Display version information and exit"
        print "  -r name        Use the given named EClass as root element"
        print "  -o             Output file name for the SMD model"
        print "If no output file is given, prints XMI file to stdout."
        print ""

    def version():
        import coral
        print "This is Coral version", coral.version, " (ecore2smd script)"
        
    import sys
    import getopt

    try:
        opts, args = getopt.getopt(arguments, "hr:o:", ["help", "version"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    opt_o = ""
    root_class = ""
    
    for o,a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif o in ("--version"):
            version()
            sys.exit(0)
        elif o in ("-r"):
            root_class = a
        elif o in ("-o"):
            opt_o = a
        else:
            usage()
            sys.exit(2)

    if not args:
        usage()
        sys.exit(2)

    import coral
    import sys
    from coral.modeler.diagram.di_support import getAllLinkedProjects

    ecore = coral.lang.ecore10
    ecore_model = coral.loadDocument("coral/lang/ecore10/def/ecore_model.xmi").contents()[0]

    ecore_model_skip_these_datatypes = (
        ecore_model.EBooleanObject,
        ecore_model.EByte,
        ecore_model.EByteArray,
        ecore_model.EByteObject,
        ecore_model.EChar,
        ecore_model.ECharacterObject,
        ecore_model.EDate,
        ecore_model.EDiagnosticChain,
        ecore_model.EDoubleObject,
        ecore_model.EEList,
        ecore_model.EEnumerator,
        ecore_model.EFeatureMap,
        ecore_model.EFeatureMapEntry,
        ecore_model.EFloatObject,
        ecore_model.EIntegerObject,
        ecore_model.EMap,
        ecore_model.EResource,
        ecore_model.EResourceSet,
        ecore_model.EShortObject,
        ecore_model.ETreeIterator,
        )        

    smd = coral.lang.SMD01

    core = coral.loadDocument("coral/core/CORE01.xmi").contents()[0]

    e2s = {}
    project = coral.loadDocument(args[0])
    allprojects = getAllLinkedProjects([project])
    
    root_class_in_smd = None
    for prj in allprojects:
        for root in prj.contents():
            for e in root.__getAllParts__():
                s = create(e)
                if s:
                    e2s[e] = s
                    if root_class and e.oclIsKindOf(ecore.EClass) and e.name == root_class:
                        root_class_in_smd = s
    
    connect(e2s)
    
    p = coral.Project()
    for e, s in e2s.iteritems():
        if e.oclIsKindOf(ecore.EPackage) and e.__getRootsProject__() == project:
            s.__setProject__(p)
            if root_class_in_smd:
                s.root_ed = root_class_in_smd
    if opt_o:
        p.setFilename(opt_o)
        coral.saveDocument(p)
    else:
        s = coral.saveDocument_str(p.contents(), "XMI 2.0")
        print s

if __name__ == "__main__":
    #
    # Set the user's locale here. Since Coral can be used as a library,
    # we do not tamper with the locale at all except at the
    # __name__ == "__main__" situations
    # The filesystem encoding is set automatically by Python
    # (and it cannot be changed after startup)
    #
    import locale
    locale.setlocale(locale.LC_ALL, '')

    import sys
    main(sys.argv[1:])
