#!/usr/bin/env python
# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.
#
# -*- coding: UTF-8 -*-
#
# io - script to display io features of Coral
#
# SYNOPSIS
#
# Explained in usage()
#
# DESCRIPTION
#
# io is a simple script that can display
# - possible input file formats
# - possible output file formats
# - possible metamodels
#
# BUGS
#
# - Doesn't recognize a difference between
# input or output file formats. They'll return the same answer.
# - Doesn't recognize if a file format is
# defective in some way, e.g., can't serialize all
# inputs
#
# AUTHORS
#
# Marcus Alanen <maalanen@abo.fi> 2005
#

def countElements(element):
    """Counts the amount of elements in the given cone element.
    """
    return element.__getAllParts__().size()
   
def main(arguments):
    def usage():
        print "Usage: countelements [ --help | --version | [ -i | -o | -m ] ]"
        print "Options:"
        print "  --help         Display this information"
        print "  --version      Display version information and exit"
        print "  -i             Display input file formats"
        print "  -o             Display output file formats"
        print "  -m             Display metamodels available"
        print "If no options are given, pretty-prints all the information."
        print ""

    def version():
        import coral
        print "This is Coral version", coral.version, " (io script)"
        
    import sys
    import getopt

    try:
        opts, args = getopt.getopt(arguments, "hiom", ["help", "version"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    opt_i = False
    opt_o = False
    opt_m = False
    
    for o,a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif o in ("--version"):
            version()
            sys.exit(0)
        elif o in ("-i"):
            opt_i = True
        elif o in ("-o"):
            opt_o = True
        elif o in ("-m"):
            opt_m = True
        else:
            usage()
            sys.exit(2)

    pretty_print = False
    
    if not (opt_i or opt_o or opt_m):
        pretty_print = True
        opt_i = True
        opt_o = True
        opt_m = True

    import coral
    IO = coral.getSupportedIOFormats()
    I = []
    O = []
    M = []
    if opt_i:
        I = list(IO)
        I.remove("GXL 1.0")
    if opt_o:
        O = list(IO)
    if opt_m:
        n = coral.xmlns_db_size()
        for i in xrange(n):
            name = coral.xmlns_db_metamodelname(i)
            ver  = coral.xmlns_db_version(i)
            rev  = coral.xmlns_db_revision(i)
            xmlns = coral.xmlns_db_xmlns(i)
            M.append( (name, ver, rev, xmlns) )

    if pretty_print:
        print "Coral version %s (built %s)" % (coral.version, coral.build_date)
        print
        print "Input Formats"
        print "-------------"
        for i in I:
            print i
        print
        print "Output Formats"
        print "--------------"
        for o in O:
            print o
        print
        
        print "Metamodels"
        print "----------"
        M = [ (m[0] + unicode(m[1]) + unicode(m[2]), m[3]) for m in M ]
        maxlen = max( [ len(m[0]) for m in M ] )
        maxlen += 1
        M.sort(lambda x,y: cmp(x[0], y[0]) )
        for m in M:
            fmt = "%" + unicode(maxlen) + "s - %s"
            print fmt % (m[0], m[1])
        print
        return
    if opt_i:
        for i in I:
            print i
    if opt_o:
        for o in O:
            print o 
    if opt_m:
        M = [ (m[0] + unicode(m[1]) + unicode(m[2]), m[3]) for m in M ]
        M.sort(lambda x,y: cmp(x[0], y[0]) )
        for m in M:
            fmt = "%s - %s"
            print fmt % (m[0], m[1])
   
if __name__ == "__main__":
    #
    # Set the user's locale here. Since Coral can be used as a library,
    # we do not tamper with the locale at all except at the
    # __name__ == "__main__" situations
    # The filesystem encoding is set automatically by Python
    # (and it cannot be changed after startup)
    #
    import locale
    locale.setlocale(locale.LC_ALL, '')

    import sys
    main(sys.argv[1:])
